#!/usr/bin/env python

import logging
import optparse
import os
import string
import sys

import hpfeeds


def log(msg):
    print('[feedcli] {0}'.format(msg))


def on_message(ident, chan, payload):
    if [i for i in payload[:20] if i not in string.printable.encode('utf-8')]:
        payload = payload[:20].encode('hex') + '...'
    log('publish to {0} by {1}: {2}'.format(chan, ident, payload))


def _main(opts, action, pubdata=None):
    try:
        hpc = hpfeeds.new(
            opts.host,
            opts.port,
            opts.ident,
            opts.secret,
            certfile=opts.certfile,
        )
    except hpfeeds.FeedException as e:
        log('Error: {0}'.format(e))
        return 1

    def on_error(payload):
        log('Error message from broker: {0}'.format(payload))
        hpc.stop()

    log('connected to {0}'.format(hpc.brokername))

    if action == 'subscribe':
        hpc.subscribe(opts.channels)
        try:
            hpc.run(on_message, on_error)
        except hpfeeds.FeedException as e:
            log('Error: {0}'.format(e))
            return 1

    elif action == 'publish':
        hpc.publish(opts.channels, pubdata)
        emsg = hpc.wait()
        if emsg:
            print('got error from server:', emsg)

    elif action == 'sendfile':
        pubfile = open(pubdata, 'rb').read()
        hpc.publish(opts.channels, pubfile)

    log('closing connection.')
    hpc.close()

    return 0


def get_parser():
    parser = optparse.OptionParser()
    parser.add_option(
        "-c", "--chan",
        action="append", dest='channels', nargs=1, type='string',
        help="channel (can be used multiple times)")
    parser.add_option(
        "-i", "--ident",
        action="store", dest='ident', nargs=1, type='string',
        help="authkey identifier")
    parser.add_option(
        "-s", "--secret",
        action="store", dest='secret', nargs=1, type='string',
        help="authkey secret")
    parser.add_option(
        "--host",
        action="store", dest='host', nargs=1, type='string',
        help="broker host")
    parser.add_option(
        "-p", "--port",
        action="store", dest='port', nargs=1, type='int',
        help="broker port")
    parser.add_option(
        "-o", "--output",
        action="store", dest='output', nargs=1, type='string',
        help="publish log filename")
    parser.add_option(
        "--tlscert",
        action="store", dest='certfile', nargs=1, type='string',
        help="certfile for ssl verification (CA)", default=None)
    parser.add_option(
        "--debug",
        action="store_const", dest='debug',
        help="enable debug log output", default=False, const=True)

    return parser


def opts():
    parser = get_parser()

    options, args = parser.parse_args()

    if len(args) < 1:
        parser.error('You need to give "subscribe" or "publish" as <action>.')
    if args[0] not in ['subscribe', 'publish', 'sendfile']:
        parser.error('You need to give "subscribe" or "publish" as <action>.')

    logging.basicConfig(level=logging.DEBUG if options.debug else logging.CRITICAL)

    action = args[0]
    data = None
    if action == 'publish':
        data = ' '.join(args[1:])
    elif action == 'sendfile':
        data = ' '.join(args[1:])

    if options.certfile:
        if not os.path.exists(options.certfile):
            parser.error('If --tlscert is specified it must be a valid path to a file')

    if not options.host:
        parser.error('You must specify a host')

    if not options.port:
        parser.error('You must specify a port')

    if not options.ident:
        parser.error('You must specify an ident')

    if not options.secret:
        parser.error('You must specify a secret')

    return options, action, data


def main():
    options, action, data = opts()
    try:
        sys.exit(_main(options, action, pubdata=data))
    except KeyboardInterrupt:
        sys.exit(0)



import asyncio
import logging
import socket
import unittest

from twisted.internet import asyncioreactor

from hpfeeds.broker import prometheus
from hpfeeds.broker.auth.memory import Authenticator
from hpfeeds.broker.server import Server
from hpfeeds.twisted import ClientSessionService


class TestClientIntegration(unittest.TestCase):

    log = logging.getLogger('hpfeeds.testserver')

    @classmethod
    def setUpClass(cls):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        asyncioreactor.install(eventloop=loop)
        cls.loop = asyncio.get_event_loop()

    def setUp(self):
        prometheus.reset()

        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_client_connections') == 0
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_connection_made') == 0
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_connection_ready', {'ident': 'test'}) is None

        self.sock = sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('127.0.0.1', 0))
        self.port = sock.getsockname()[1]

        authenticator = Authenticator({
            'test': {
                'secret': 'secret',
                'subchans': ['test-chan'],
                'pubchans': ['test-chan'],
                'owner': 'some-owner',
            }
        })

        self.server = Server(authenticator, sock=self.sock)

    def test_subscribe_and_publish(self):
        async def inner():
            self.log.debug('Starting server')
            server_future = asyncio.ensure_future(self.server.serve_forever())

            self.log.debug('Creating client service')
            client = ClientSessionService(f'tcp:127.0.0.1:{self.port}', 'test', 'secret')
            client.subscribe('test-chan')
            client.startService()

            # Wait till client connected
            await client.whenConnected.asFuture(self.loop)

            assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_client_connections') == 1
            assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_connection_made') == 1

            self.log.debug('Publishing test message')
            client.publish('test-chan', b'test message')

            self.log.debug('Waiting for read()')
            assert ('test', 'test-chan', b'test message') == await client.read().asFuture(self.loop)

            # We would test this after call to subscribe, but need to wait until sure server has processed command
            assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_subscriptions', {'ident': 'test', 'chan': 'test-chan'}) == 1

            # This will only have incremented when server has processed auth message
            # Test can only reliably assert this is the case after reading a message
            assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_connection_ready', {'ident': 'test'}) == 1

            self.log.debug('Stopping client')
            await client.stopService().asFuture(self.loop)

            self.log.debug('Stopping server')
            server_future.cancel()
            await server_future

        self.loop.run_until_complete(inner())
        assert len(self.server.connections) == 0, 'Connection left dangling'
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_client_connections') == 0
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_connection_lost', {'ident': 'test'}) == 1

        # Closing should auto unsubscribe
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_subscriptions', {'ident': 'test', 'chan': 'test-chan'}) == 0

    def test_late_subscribe_and_publish(self):
        async def inner():
            self.log.debug('Starting server')
            server_future = asyncio.ensure_future(self.server.serve_forever())

            self.log.debug('Creating client service')
            client = ClientSessionService(f'tcp:127.0.0.1:{self.port}', 'test', 'secret')
            client.startService()

            # Wait till client connected
            await client.whenConnected.asFuture(self.loop)

            assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_client_connections') == 1
            assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_connection_made') == 1

            # Subscribe to a new thing after connection is up
            client.subscribe('test-chan')

            self.log.debug('Publishing test message')
            client.publish('test-chan', b'test message')

            self.log.debug('Waiting for read()')
            assert ('test', 'test-chan', b'test message') == await client.read().asFuture(self.loop)

            # We would test this after call to subscribe, but need to wait until sure server has processed command
            assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_subscriptions', {'ident': 'test', 'chan': 'test-chan'}) == 1

            # Unsubscribe while the connection is up
            client.unsubscribe('test-chan')

            # This will only have incremented when server has processed auth message
            # Test can only reliably assert this is the case after reading a message
            assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_connection_ready', {'ident': 'test'}) == 1

            self.log.debug('Stopping client')
            await client.stopService().asFuture(self.loop)

            self.log.debug('Stopping server')
            server_future.cancel()
            await server_future

        self.loop.run_until_complete(inner())
        assert len(self.server.connections) == 0, 'Connection left dangling'
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_client_connections') == 0
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_connection_lost', {'ident': 'test'}) == 1

        # Again, we should test this directly after calling unsubscribe(), but no ability to wait
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_subscriptions', {'ident': 'test', 'chan': 'test-chan'}) == 0



from twisted.internet.protocol import Factory

from .protocol import ClientProtocol


class ClientFactory(Factory):

    protocol = ClientProtocol

    def __init__(self, ident, secret):
        self.ident = ident
        self.secret = secret



from twisted.internet.protocol import Protocol
from twisted.python import log

from hpfeeds.exceptions import ProtocolException
from hpfeeds.protocol import (
    OP_AUTH,
    OP_ERROR,
    OP_INFO,
    OP_PUBLISH,
    OP_SUBSCRIBE,
    OP_UNSUBSCRIBE,
    Unpacker,
    msgauth,
    msgerror,
    msginfo,
    msgpublish,
    msgsubscribe,
    msgunsubscribe,
    readauth,
    readerror,
    readinfo,
    readpublish,
    readsubscribe,
    readunsubscribe,
)


class BaseProtocol(Protocol):

    def __init__(self):
        self.unpacker = Unpacker()

    def protocolError(self, reason):
        '''
        Called when an unrecoverable protocol error has been detected. The
        connection will be dropped.
        '''
        log.err(reason)

    def onError(self, error):
        '''
        Called by messageReceived when an OP_ERROR has been parsed.
        '''
        raise NotImplementedError(self.onError)

    def onInfo(self, name, rand):
        '''
        Called by messageReceived when an OP_INFO has been parsed.
        '''
        raise NotImplementedError(self.onInfo)

    def onAuth(self, ident, secret):
        '''
        Called by messageReceived when an OP_AUTH has been parsed.
        '''
        raise NotImplementedError(self.onAuth)

    def onPublish(self, ident, chan, data):
        '''
        Called by messageReceived when an OP_PUBLISH has been parsed.
        '''
        raise NotImplementedError(self.onPublish)

    def onSubscribe(self, ident, chan):
        '''
        Called by messageReceived when an OP_SUBSCRIBE has been parsed.
        '''
        raise NotImplementedError(self.onSubscribe)

    def onUnsubscribe(self, ident, chan):
        '''
        Called by messageReceived when an OP_UNSUBSCRIBE has been parsed.
        '''
        raise NotImplementedError(self.onUnsubscribe)

    def messageReceived(self, opcode, data):
        if opcode == OP_ERROR:
            return self.onError(readerror(data))
        elif opcode == OP_INFO:
            return self.onInfo(*readinfo(data))
        elif opcode == OP_AUTH:
            return self.onAuth(*readauth(data))
        elif opcode == OP_PUBLISH:
            return self.onPublish(*readpublish(data))
        elif opcode == OP_SUBSCRIBE:
            return self.onSubscribe(*readsubscribe(data))
        elif opcode == OP_UNSUBSCRIBE:
            return self.onUnsubscribe(*readunsubscribe(data))

        # Can't recover from an unknown opcode, so drop connection
        self.protocolError('Unknown message opcode: {!r}'.format(opcode))
        self.transport.loseConnection()

    def dataReceived(self, data):
        self.unpacker.feed(data)
        try:
            for opcode, data in self.unpacker:
                self.messageReceived(opcode, data)
        except ProtocolException as e:
            # Can't recover from a protocol decoding error, so drop connection
            self.protocolError(str(e))
            self.transport.loseConnection()

    def error(self, error):
        self.transport.write(msgerror(error))

    def info(self, name, rand):
        self.transport.write(msginfo(name, rand))

    def auth(self, rand, ident, secret):
        self.transport.write(msgauth(rand, ident, secret))

    def publish(self, ident, channel, payload):
        self.transport.write(msgpublish(ident, channel, payload))

    def subscribe(self, ident, channel):
        self.transport.write(msgsubscribe(ident, channel))

    def unsubscribe(self, ident, channel):
        self.transport.write(msgunsubscribe(ident, channel))


class ClientProtocol(BaseProtocol):

    '''
    A base class for implementing hpfeeds client functionality in Twisted.
    '''

    def connectionReady(self):
        '''
        Called when a connection has been established and authentication has
        completed.
        '''
        pass

    def onInfo(self, name, rand):
        '''
        Called by messageReceived when an OP_INFO has been parsed.

        The default client implementation will send an appropriate OP_AUTH and
        call connectionReady.
        '''
        self.auth(rand, self.factory.ident, self.factory.secret)
        self.connectionReady()

    def onAuth(self, ident, secret):
        '''
        Called by messageReceived when an OP_AUTH has been parsed.

        A HPFeeds **client** should never receive an OP_AUTH. By default it will
        call protocolError and drop the connection.
        '''
        self.protocolError('Unexpected OP_AUTH')
        self.transport.loseConnection()

    def onSubscribe(self, ident, chan):
        '''
        Called by messageReceived when an OP_SUBSCRIBE has been parsed.

        A HPFeeds **client** should never receive an OP_SUBSCRIBE. By default
        it will call protocolError and drop the connection.
        '''
        self.protocolError('Unexpected OP_SUBSCRIBE')
        self.transport.loseConnection()

    def onUnsubscribe(self, ident, chan):
        '''
        Called by messageReceived when an OP_UNSUBSCRIBE has been parsed.

        A HPFeeds **client** should never receive an OP_UNSUBSCRIBE. By default
        it will call protocolError and drop the connection.
        '''
        self.protocolError('Unexpected OP_UNSUBSCRIBE')
        self.transport.loseConnection()

    def error(self, error):
        raise RuntimeError('Client tried to set OP_ERROR to server, this is a protocol violation')

    def info(self, name, rand):
        raise RuntimeError('Client tried to set OP_INFO to server, this is a protocol violation')



import asyncio
import logging
import socket
import sys

from .protocol import ClientProtocol


class _Protocol(ClientProtocol):

    '''
    An implementation of the HPFeeds AIO Client Protocol that is used
    by Client.
    '''

    def __init__(self, client):
        self.client = client
        super().__init__(client.ident, client.secret)

    def connection_made(self, transport):
        sock = transport.get_extra_info('socket')
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if sys.platform.startswith('linux'):
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 10)
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 5)
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 3)
        super().connection_made(transport)

    def connection_ready(self):
        '''
        Called when a connection has been established and authentication has
        completed.

        Subscribe to any requested channels.
        '''
        self.client.protocol = self

        for topic in self.client.subscriptions:
            self.subscribe(self.ident, topic)

        # We are connected and ready for business
        self.client.when_connected.set_result(None)

    def connection_lost(self, reason):
        self.client.when_closed.set_result(None)

    def on_publish(self, ident, chan, data):
        '''
        Called by messageReceived when an OP_PUBLISH has been parsed.

        All received messages are pushed into owning services read_queue (a
        DeferredQueue).
        '''
        self.client.read_queue.put_nowait((ident, chan, data))


class ClientSession(object):

    '''
    A service that maintains a connection to a hpfeeds broker and provides
    helpers for reading and writing to the broker.
    '''

    log = logging.getLogger('hpfeeds.asyncio.client')

    def __init__(self, host, port, ident, secret):
        self.host = host
        self.port = port
        self.ident = ident
        self.secret = secret

        self.read_queue = asyncio.Queue()
        self.subscriptions = set()
        self.protocol = None

        self.when_connected = asyncio.Future()
        self.closing = False
        self._ensure_connected = asyncio.ensure_future(self.reconnect())

    async def _tryconnect(self):
        while True:
            try:
                client = await asyncio.get_event_loop().create_connection(
                    lambda: _Protocol(self),
                    self.host,
                    self.port,
                )
                return client
            except OSError as e:
                pass

            await asyncio.sleep(1)

    async def reconnect(self):
        while not self.closing:
            self.protocol = None
            self.when_closed = asyncio.Future()

            try:
                await self._tryconnect()
            except Exception as e:
                self. log.exception(e)

            await self.when_closed
            self.when_connected = asyncio.Future()

    def publish(self, chan, payload):
        if self.protocol:
            self.protocol.publish(self.ident, chan, payload)

    def subscribe(self, topic):
        if topic not in self.subscriptions:
            self.subscriptions.add(topic)
            if self.protocol:
                self.protocol.subscribe(self.ident, topic)

    def unsubscribe(self, topic):
        if topic in self.subscriptions:
            self.subscriptions.discard(topic)
            if self.protocol:
                self.protocol.unsubscribe(self.ident, topic)

    def read(self):
        '''
        Returns a future which fires with the next available message received
        from the broker.

        If a message has already been received and is queued the future will
        fire immediately.
        '''
        return self.read_queue.get()

    async def close(self):
        self.closing = True
        if self.protocol:
            self.protocol.transport.close()
        await self.when_closed

    def __enter__(self):
        raise TypeError("Use async with instead")

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    async def __aenter__(self):
        await self.when_connected
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



import unittest

from hpfeeds.protocol import (
    Unpacker,
    msgauth,
    msghdr,
    msgpublish,
    msgsubscribe,
    readauth,
    readinfo,
    readpublish,
    readsubscribe,
)


class TestMessageBuilder(unittest.TestCase):

    def test_msghdr(self):
        assert msghdr(1, b'abcdefg') == b'\x00\x00\x00\x0c\x01abcdefg'

    def test_msgpublish(self):
        msg = msgpublish('ident', 'chan', 'somedata')
        assert msg == b'\x00\x00\x00\x18\x03\x05ident\x04chansomedata'

    def test_msgsubscribe(self):
        msg = msgsubscribe('ident', 'chan')
        assert msg == b'\x00\x00\x00\x0f\x04\x05identchan'

    def test_msgauth(self):
        msg = msgauth(b'rand', 'ident', 'secret')
        assert msg == (
            b'\x00\x00\x00\x1f\x02\x05ident\xbf\xa9^\x11I\xcd\x9es'
            b'\x80\xfd\xfcaJW\tZ\xb7\x19\xc1\xb4'
        )


class TestMessageReader(unittest.TestCase):

    def test_readinfo(self):
        name, rand = readinfo(b'\x07hpfeeds\x01 a\xff')
        assert name == 'hpfeeds'
        assert rand == b'\x01 a\xff'

    def test_readauth(self):
        ident, secret = readauth(
            b'\x05ident\xbf\xa9^\x11I\xcd\x9es'
            b'\x80\xfd\xfcaJW\tZ\xb7\x19\xc1\xb4'
        )
        assert ident == 'ident'
        assert secret == (
            b'\xbf\xa9^\x11I\xcd\x9es\x80\xfd\xfcaJW\tZ\xb7\x19\xc1\xb4'
        )

    def test_readpublish(self):
        ident, chan, data = readpublish(b'\x05ident\x04chansomedata')
        assert ident == 'ident'
        assert chan == 'chan'
        assert data == b'somedata'

    def test_readsubscribe(self):
        ident, chan = readsubscribe(b'\x05identchan')
        assert ident == 'ident'
        assert chan == 'chan'


class TestUnpacker(unittest.TestCase):

    def test_unpack_at_max_message_size_unicode_1(self):
        unpacker = Unpacker()
        data = msgpublish(b'b', b'a', '\u2603'.encode('utf-8') * 349524)
        unpacker.feed(data)
        packets = list(iter(unpacker))
        assert len(unpacker.buf) == 0
        assert packets[0][0] == 3
        assert len(packets[0][1]) == (1024 ** 2)
        # assert packets[0][1].endswith(b'a' * 349525)

    def test_unpack_at_max_message_size_unicode_2(self):
        unpacker = Unpacker()
        unpacker.feed(msgpublish(b'', b'', '\u001F'.encode('utf-8') * 1048574))
        packets = list(iter(unpacker))
        assert len(unpacker.buf) == 0
        assert packets[0][0] == 3
        assert len(packets[0][1]) == (1024 ** 2)
        # assert packets[0][1].endswith(b'a' * 349525)

    def test_unpack_at_max_message_size(self):
        unpacker = Unpacker()
        unpacker.feed(msgpublish(b'', b'', b'a' * 1048574))
        packets = list(iter(unpacker))
        assert len(unpacker.buf) == 0
        assert packets[0][0] == 3
        assert len(packets[0][1]) == (1024 ** 2)
        assert packets[0][1].endswith(b'a' * 1048574)

    def test_unpack_at_max_message_size_leftovers(self):
        unpacker = Unpacker()
        unpacker.feed(msgpublish(b'', b'', b'a' * 1048574) + b'z' * 1024)
        packets = [next(iter(unpacker))]
        assert len(unpacker.buf) == 1024
        assert packets[0][0] == 3
        # assert len(packets[0][1]) == (1024 ** 2)
        # assert packets[0][1].endswith(b'a' * 1048574)

    def test_unpack_1(self):
        unpacker = Unpacker()
        unpacker.feed(msghdr(1, b'abcdefghijklmnopqrstuvwxyz'))
        packets = list(iter(unpacker))
        assert packets == [(1, b'abcdefghijklmnopqrstuvwxyz')]

    def test_unpack_2(self):
        message = msghdr(1, b'abcdefghijklmnopqrstuvwxyz')
        unpacker = Unpacker()

        # The unpacker shouldn't yield any messages until it has consumed the
        # full object
        for b in message[:-1]:
            unpacker.feed([b])
            assert list(iter(unpacker)) == []

        unpacker.feed([message[-1]])
        assert list(iter(unpacker)) == [(1, b'abcdefghijklmnopqrstuvwxyz')]



import unittest

from hpfeeds.broker.auth.sqlite import Authenticator


class TestAuthenticator(unittest.TestCase):

    def test_(self):
        a = Authenticator(':memory:')
        a.sql.execute(
            'INSERT INTO authkeys(owner, ident, secret, pubchans, subchans)'
            'VALUES ("a", "b", "c", "[]", "[]")'
        )
        key = a.get_authkey('b')
        assert key['owner'] == 'a'
        assert key['secret'] == 'c'
        assert key['pubchans'] == []
        assert key['subchans'] == []



import asyncio

from hpfeeds.exceptions import ProtocolException
from hpfeeds.protocol import (
    OP_AUTH,
    OP_ERROR,
    OP_INFO,
    OP_PUBLISH,
    OP_SUBSCRIBE,
    OP_UNSUBSCRIBE,
    Unpacker,
    msgauth,
    msgerror,
    msginfo,
    msgpublish,
    msgsubscribe,
    msgunsubscribe,
    readauth,
    readerror,
    readinfo,
    readpublish,
    readsubscribe,
    readunsubscribe,
)


class BaseProtocol(asyncio.Protocol):

    def __init__(self):
        self.unpacker = Unpacker()

    def protocol_error(self, reason):
        '''
        Called when an unrecoverable protocol error has been detected. The
        connection will be dropped.
        '''
        pass

    def on_error(self, error):
        '''
        Called by message_received when an OP_ERROR has been parsed.
        '''
        raise NotImplementedError(self.on_error)

    def on_info(self, name, rand):
        '''
        Called by message_received when an OP_INFO has been parsed.
        '''
        raise NotImplementedError(self.on_info)

    def on_auth(self, ident, secret):
        '''
        Called by message_received when an OP_AUTH has been parsed.
        '''
        raise NotImplementedError(self.on_auth)

    def on_publish(self, ident, chan, data):
        '''
        Called by message_received when an OP_PUBLISH has been parsed.
        '''
        raise NotImplementedError(self.on_publish)

    def on_subscribe(self, ident, chan):
        '''
        Called by message_received when an OP_SUBSCRIBE has been parsed.
        '''
        raise NotImplementedError(self.on_subscribe)

    def on_unsubscribe(self, ident, chan):
        '''
        Called by message_received when an OP_UNSUBSCRIBE has been parsed.
        '''
        raise NotImplementedError(self.on_unsubscribe)

    def message_received(self, opcode, data):
        if opcode == OP_ERROR:
            return self.on_error(readerror(data))
        elif opcode == OP_INFO:
            return self.on_info(*readinfo(data))
        elif opcode == OP_AUTH:
            return self.on_auth(*readauth(data))
        elif opcode == OP_PUBLISH:
            return self.on_publish(*readpublish(data))
        elif opcode == OP_SUBSCRIBE:
            return self.on_subscribe(*readsubscribe(data))
        elif opcode == OP_UNSUBSCRIBE:
            return self.on_unsubscribe(*readunsubscribe(data))

        # Can't recover from an unknown opcode, so drop connection
        self.protocol_error('Unknown message opcode: {!r}'.format(opcode))
        self.transport.close()

    def connection_made(self, transport):
        self.transport = transport

    def data_received(self, data):
        self.unpacker.feed(data)
        try:
            for opcode, data in self.unpacker:
                self.message_received(opcode, data)
        except ProtocolException as e:
            # Can't recover from a protocol decoding error, so drop connection
            self.protocol_error(str(e))
            self.transport.close()

    def error(self, error):
        self.transport.write(msgerror(error))

    def info(self, name, rand):
        self.transport.write(msginfo(name, rand))

    def auth(self, rand, ident, secret):
        self.transport.write(msgauth(rand, ident, secret))

    def publish(self, ident, channel, payload):
        self.transport.write(msgpublish(ident, channel, payload))

    def subscribe(self, ident, channel):
        self.transport.write(msgsubscribe(ident, channel))

    def unsubscribe(self, ident, channel):
        self.transport.write(msgunsubscribe(ident, channel))


class ClientProtocol(BaseProtocol):

    '''
    A base class for implementing hpfeeds client functionality in aio.
    '''

    def __init__(self, ident, secret):
        self.ident = ident
        self.secret = secret
        super().__init__()

    def connection_ready(self):
        '''
        Called when a connection has been established and authentication has
        completed.
        '''
        pass

    def on_info(self, name, rand):
        '''
        Called by message_received when an OP_INFO has been parsed.

        The default client implementation will send an appropriate OP_AUTH and
        call connection_ready.
        '''
        self.auth(rand, self.ident, self.secret)
        self.connection_ready()

    def on_auth(self, ident, secret):
        '''
        Called by message_received when an OP_AUTH has been parsed.

        A HPFeeds **client** should never receive an OP_AUTH. By default it will
        call protocol_error and drop the connection.
        '''
        self.protocol_error('Unexpected OP_AUTH')
        self.transport.close()

    def on_subscribe(self, ident, chan):
        '''
        Called by message_received when an OP_SUBSCRIBE has been parsed.

        A HPFeeds **client** should never receive an OP_SUBSCRIBE. By default
        it will call protocol_error and drop the connection.
        '''
        self.protocol_error('Unexpected OP_SUBSCRIBE')
        self.transport.close()

    def on_unsubscribe(self, ident, chan):
        '''
        Called by message_received when an OP_UNSUBSCRIBE has been parsed.

        A HPFeeds **client** should never receive an OP_UNSUBSCRIBE. By default
        it will call protocol_error and drop the connection.
        '''
        self.protocol_error('Unexpected OP_UNSUBSCRIBE')
        self.transport.close()

    def error(self, error):
        raise RuntimeError('Client tried to set OP_ERROR to server, this is a protocol violation')

    def info(self, name, rand):
        raise RuntimeError('Client tried to set OP_INFO to server, this is a protocol violation')



import argparse
import logging
import ssl

import aiorun

from hpfeeds.broker.auth import env, sqlite
from hpfeeds.broker.server import Server


def main():
    parser = argparse.ArgumentParser(description='Run a hpfeeds broker')
    parser.add_argument('--bind', default='0.0.0.0:20000', action='store')
    parser.add_argument('--exporter', default='', action='store')
    parser.add_argument('--name', default='hpfeeds', action='store')
    parser.add_argument('--debug', default=False, action='store_true')
    parser.add_argument('--auth', default='sqlite', action='store')
    parser.add_argument('--tlscert', default=None, action='store')
    parser.add_argument('--tlskey', default=None, action='store')
    args = parser.parse_args()

    if (args.tlscert and not args.tlskey) or (args.tlskey and not args.tlscert):
        parser.error('Must specify --tlskey AND --tlscert')
        return

    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO,
    )

    if args.auth == 'env':
        auth = env.Authenticator()
    else:
        auth = sqlite.Authenticator('sqlite.db')

    ssl_context = None
    if args.tlscert:
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(args.tlscert, args.tlskey)

    broker = Server(
        auth=auth,
        bind=args.bind,
        exporter=args.exporter,
        name=args.name,
        ssl=ssl_context,
    )

    return aiorun.run(broker.serve_forever())


if __name__ == '__main__':
    main()



import os


def get_key(ident, value, default=None):
    return os.environ.get(f'HPFEEDS_{ident.upper()}_{value.upper()}', default=default)


class Authenticator(object):

    '''
    Authentication class that takes a mapping of user identities and
    their permissions.

    Given environment variables:

        HPFEEDS_IDENT1_SECRET=secret1
        HPFEEDS_IDENT1_SUBCHANS=test
        HPFEEDS_IDENT1_PUBCHANS=test

    A user called `ident1` will be able to authenticate with the broker.

    authenticator = Authenticator()
    '''

    def close(self):
        pass

    def get_authkey(self, ident):
        secret = get_key(ident, 'secret')
        if not secret:
            return None

        return {
            'ident': ident,
            'owner': get_key(ident, 'owner', ident),
            'secret': secret,
            'subchans': get_key(ident, 'subchans', '').split(','),
            'pubchans': get_key(ident, 'pubchans', '').split(','),
        }



class Authenticator(object):

    '''
    Authentication class that takes a mapping of user identities and
    their permissions.

    authenticator = Authenticator({
        'ident1': {
            'secret': 'somesecret',
            'pubchans': ['channel1'],
            'subchans': ['channel2'],
            'owner': 'youruser',
        }
    })
    '''

    def __init__(self, creds):
        self.creds = creds

    def close(self):
        pass

    def get_authkey(self, ident):
        authkey = self.creds.get(ident, None)
        if not authkey:
            return

        authkey = dict(authkey)
        authkey['ident'] = ident
        return authkey



from .factory import ClientFactory
from .protocol import BaseProtocol, ClientProtocol
from .service import ClientSessionService


__all__ = [
    'BaseProtocol',
    'ClientFactory',
    'ClientProtocol',
    'ClientSessionService',
]



# Copyright (C) 2010-2013 Mark Schloesser <ms@mwcollect.org
# This file is part of hpfeeds - https://github.com/rep/hpfeeds
# See the file 'LICENSE' for copying permission.

import logging
import socket
import ssl
import sys
import threading
import time

from .exceptions import Disconnect, FeedException
from .protocol import (
    BUFSIZ,
    OP_ERROR,
    OP_INFO,
    OP_PUBLISH,
    Unpacker,
    msgauth,
    msgpublish,
    msgsubscribe,
    readerror,
    readinfo,
    readpublish,
)

logger = logging.getLogger('pyhpfeeds')


__all__ = ["new", "FeedException"]


class Client(object):

    def __init__(self, host, port, ident, secret, timeout=3, reconnect=True, sleepwait=20):
        self.host, self.port = host, port
        self.ident, self.secret = ident, secret
        self.timeout = timeout
        self.reconnect = reconnect
        self.sleepwait = sleepwait
        self.brokername = 'unknown'
        self.connected = False
        self.stopped = False
        self.s = None
        self.connecting_lock = threading.Lock()
        self.subscriptions = set()
        self.unpacker = Unpacker()

        self.tryconnect()

    def makesocket(self, addr_family):
        return socket.socket(addr_family, socket.SOCK_STREAM)

    def recv(self):
        try:
            d = self.s.recv(BUFSIZ)
        except socket.timeout:
            return ""
        except socket.error as e:
            logger.warn("Socket error: %s", e)
            raise Disconnect()

        if not d:
            logger.warn("recv() returned empty string")
            raise Disconnect()

        return d

    def send(self, data):
        try:
            self.s.sendall(data)
        except socket.timeout:
            logger.warn("Timeout while sending - disconnect.")
            raise Disconnect()
        except socket.error as e:
            logger.warn("Socket error: %s", e)
            raise Disconnect()

        return True

    def tryconnect(self):
        with self.connecting_lock:
            if self.connected:
                return

            while True:
                try:
                    return self.connect()
                except socket.error as e:
                    logger.warn(
                        'Socket error while connecting',
                        exc_info=e,
                    )
                except FeedException as e:
                    logger.warn(
                        'FeedException while connecting',
                        exc_info=e,
                    )
                except Disconnect as e:
                    logger.warn('Disconnect while connecting.')

                time.sleep(self.sleepwait)

    def connect(self):
        self.close_old()

        logger.info('connecting to {0}:{1}'.format(self.host, self.port))

        # Try other resolved addresses (IPv4 or IPv6) if failed.
        ainfos = socket.getaddrinfo(
            self.host,
            1,
            socket.AF_UNSPEC,
            socket.SOCK_STREAM,
        )

        for ainfo in ainfos:
            addr_family = ainfo[0]
            addr = ainfo[4][0]
            try:
                self.s = self.makesocket(addr_family)
                self.s.settimeout(self.timeout)
                self.s.connect((addr, self.port))
            except Exception:
                logger.exception('Could not connect to broker {}[{}]'.format(
                    self.host,
                    addr,
                ))
                continue
            else:
                self.connected = True
                break

        if self.connected is False:
            raise FeedException(
                'Could not connect to broker {}'.format(self.host)
            )

        self.unpacker.reset()

        try:
            d = self.s.recv(BUFSIZ)
        except socket.timeout:
            raise FeedException('Connection receive timeout.')

        self.unpacker.feed(d)
        for opcode, data in self.unpacker:
            if opcode == OP_INFO:
                name, rand = readinfo(data)
                logger.debug(
                    'info message name: {0}, rand: {1!r}'.format(name, rand)
                )
                self.brokername = name

                self.send(msgauth(rand, self.ident, self.secret))
                break
            else:
                raise FeedException('Expected OP_INFO but got another opcode.')
        else:
            raise FeedException('Expected OP_INFO but cannot assemble complete message')

        self.s.settimeout(None)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if sys.platform.startswith('linux'):
            self.s.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 10)

    def run(self, message_callback, error_callback):
        while not self.stopped:
            self._subscribe()
            while self.connected:
                try:
                    d = self.recv()
                    self.unpacker.feed(d)

                    for opcode, data in self.unpacker:
                        if opcode == OP_PUBLISH:
                            message_callback(*readpublish(data))
                        elif opcode == OP_ERROR:
                            error_callback(readerror(data))

                except Disconnect as e:
                    self.connected = False
                    logger.info('Disconnected from broker.', exc_info=e)
                    break

                # end run loops if stopped
                if self.stopped:
                    break

            if not self.stopped and self.reconnect:
                # connect again if disconnected
                self.tryconnect()

        logger.info('Stopped, exiting run loop.')

    def _read_message(self):
        d = self.recv()
        if not d:
            raise Disconnect()

        self.unpacker.feed(d)
        for opcode, data in self.unpacker:
            return (opcode, data)

    def wait(self, timeout=1):
        self.s.settimeout(timeout)

        try:
            d = self.recv()
            if not d:
                return None

            self.unpacker.feed(d)
            for opcode, data in self.unpacker:
                if opcode == OP_ERROR:
                    return readerror(data)
        except Disconnect:
            pass

        return None

    def close_old(self):
        if self.s:
            try:
                self.s.close()
            except Exception:
                pass

    def subscribe(self, chaninfo):
        if type(chaninfo) == str:
            chaninfo = [chaninfo]
        for c in chaninfo:
            self.subscriptions.add(c)

    def _subscribe(self):
        for c in self.subscriptions:
            try:
                logger.debug('Sending subscription for {0}.'.format(c))
                self.send(msgsubscribe(self.ident, c))
            except Disconnect:
                self.connected = False
                logger.info('Disconnected from broker (in subscribe).')
                if not self.reconnect:
                    raise
                break

    def publish(self, chaninfo, data):
        if type(chaninfo) == str:
            chaninfo = [chaninfo]
        for c in chaninfo:
            try:
                self.send(msgpublish(self.ident, c, data))
            except Disconnect:
                self.connected = False
                logger.info('Disconnected from broker (in publish).')
                if self.reconnect:
                    self.tryconnect()
                else:
                    raise

    def stop(self):
        self.stopped = True

    def close(self):
        try:
            self.s.close()
        except Exception:
            logger.debug('Socket exception when closing (ignored though).')


class SslClient(Client):
    def __init__(self, *args, **kwargs):
        self.certfile = kwargs.pop("certfile", None)
        super(SslClient, self).__init__(*args, **kwargs)

    def makesocket(self, addr_family):
        sock = socket.socket(addr_family, socket.SOCK_STREAM)
        return ssl.wrap_socket(
            sock,
            ca_certs=self.certfile,
            ssl_version=3,
            cert_reqs=2,
        )


def new(host=None, port=10000, ident=None, secret=None, timeout=3, reconnect=True, sleepwait=20, certfile=None):
    if certfile:
        return SslClient(host, port, ident, secret, timeout, reconnect, certfile=certfile)
    return Client(host, port, ident, secret, timeout, reconnect)



#!/usr/bin/python
# -*- coding: utf8 -*-

import asyncio
import collections
import logging

from .connection import Connection
from .prometheus import (
    RECEIVE_PUBLISH_COUNT,
    RECEIVE_PUBLISH_SIZE,
    SUBSCRIPTIONS,
    start_metrics_server,
)

log = logging.getLogger("hpfeeds.broker")


class Server(object):

    def __init__(self, auth, bind=None, exporter=None, sock=None, ssl=None, name='hpfeeds'):
        self.auth = auth
        self.name = name

        self.host, self.port = self._parse_endpoint(bind)
        self.sock = sock
        self.ssl = ssl

        self.exporter = self._parse_endpoint(exporter)

        self.connections = set()
        self.subscriptions = collections.defaultdict(list)

    def _parse_endpoint(self, endpoint):
        if not endpoint:
            return (None, None)
        elif ':' not in endpoint:
            raise ValueError('Invalid bind addr')
        else:
            return endpoint.split(':', 1)

    def get_authkey(self, identifier):
        return self.auth.get_authkey(identifier)

    def publish(self, source, chan, data):
        '''
        Called by a connection to push data to all subscribers of a channel
        '''
        RECEIVE_PUBLISH_COUNT.labels(source.ak, chan).inc()
        RECEIVE_PUBLISH_SIZE.labels(source.ak, chan).observe(len(data))

        for dest in self.subscriptions[chan]:
            dest.publish(source.ak, chan, data)

    def subscribe(self, source, chan):
        '''
        Subscribe a connection to a channel
        '''
        SUBSCRIPTIONS.labels(source.ak, chan).inc()
        self.subscriptions[chan].append(source)
        source.active_subscriptions.add(chan)

    def unsubscribe(self, source, chan):
        '''
        Unsubscribe a connection from a channel
        '''
        if chan in source.active_subscriptions:
            source.active_subscriptions.remove(chan)
        if source in self.subscriptions[chan]:
            self.subscriptions[chan].remove(source)
        SUBSCRIPTIONS.labels(source.ak, chan).dec()

    async def serve_forever(self):
        ''' Start handling connections. Await on this to listen forever. '''

        if self.exporter:
            metrics_server = await start_metrics_server(*self.exporter)

        server = await asyncio.get_event_loop().create_server(
            lambda: Connection(self),
            host=self.host,
            port=self.port,
            sock=self.sock,
            ssl=self.ssl,
        )

        try:
            while True:
                await asyncio.sleep(10)
        except asyncio.CancelledError:
            server.close()

            # for future in asyncio.as_completed([c.close() for c in list(self.connections)]):
            #    try:
            #        await future
            #    except Exception as e:
            #        log.exception(e)

            log.debug(f'Waiting for {self} to wrap up')
            await server.wait_closed()

            if self.exporter:
                log.debug('Waiting for stats server to wrap up')
                await metrics_server.cleanup()



import unittest

from hpfeeds.twisted import ClientFactory, ClientProtocol


class TestClientProtocol(unittest.TestCase):

    def test_instance_factory(self):
        factory = ClientFactory('ident', 'secret')
        assert factory.ident == 'ident'
        assert factory.secret == 'secret'

    def test_for_protocol(self):
        factory = ClientFactory.forProtocol(ClientProtocol, 'ident', 'secret')
        assert factory.ident == 'ident'
        assert factory.secret == 'secret'



import json
import sqlite3

# inserting a user:
# sqlite3 db.sqlite3
# > insert into authkeys (owner, ident, secret, pubchans, subchans)
#      values ('owner', 'ident', 'secret', '["chan1"]', '["chan1"]');


class Authenticator(object):

    def __init__(self, path):
        self.sql = sqlite3.connect(path)
        self.check_db()

    def check_db(self):
        with self.sql:
            try:
                self.sql.execute("select * from logs, authkeys where 1=0")
            except sqlite3.OperationalError:
                print("setting up tables...")
                # create tables
                self.sql.execute("""
                create table logs (id integer primary key autoincrement,
                    data TEXT)
                """)
                self.sql.execute("""
                create table stats (id integer primary key autoincrement,
                    ak TEXT, uid TEXT, data TEXT)
                """)
                self.sql.execute("""
                create table authkeys (id integer primary key autoincrement,
                    owner TEXT, ident TEXT, secret TEXT,
                    pubchans TEXT, subchans TEXT)
                """)

    def close(self):
        self.sql.close()

    def get_authkey(self, ident):
        c = self.sql.cursor()
        try:
            c.execute("select * from authkeys where ident=?", (ident,))
            res = c.fetchone()
        except Exception:
            import traceback
            traceback.print_exc()
            return None
        finally:
            c.close()

        if not res:
            return None

        ak, owner, ident, secret, pubchans, subchans = res

        pubchans = json.loads(pubchans)
        subchans = json.loads(subchans)

        return dict(
            secret=secret,
            ident=ident,
            pubchans=pubchans,
            subchans=subchans,
            owner=owner,
        )



from twisted.application.internet import ClientService
from twisted.application.service import MultiService
from twisted.internet import defer
from twisted.internet.endpoints import clientFromString

from .factory import ClientFactory
from .protocol import ClientProtocol


class _Protocol(ClientProtocol):

    '''
    An implementation of the HPFeeds Twisted Client Protocol that is used
    by ClientService.
    '''

    def connectionMade(self):
        if hasattr(self.transport, 'setTcpKeepAlive'):
            self.transport.setTcpKeepAlive(10)

    def connectionReady(self):
        '''
        Called when a connection has been established and authentication has
        completed.

        Subscribe to any requested channels.
        '''
        self.factory.service.protocol = self

        for topic in self.factory.service.subscriptions:
            self.subscribe(self.factory.service.ident, topic)

        # We are connected and ready for business
        self.factory.service.whenConnected.callback(None)

    def connectionLost(self, reason):
        self.factory.service.protocol = None
        self.factory.service.whenConnected = defer.Deferred()
        self.factory = None

    def onPublish(self, ident, chan, data):
        '''
        Called by messageReceived when an OP_PUBLISH has been parsed.

        All received messages are pushed into owning services read_queue (a
        DeferredQueue).
        '''
        self.factory.service.read_queue.put((ident, chan, data))


class ClientSessionService(MultiService):

    '''
    A service that maintains a connection to a hpfeeds broker and provides
    helpers for reading and writing to the broker.
    '''

    def __init__(self, endpoint, ident, secret, retryPolicy=None):
        super(ClientSessionService, self).__init__()

        self.ident = ident

        self.read_queue = defer.DeferredQueue()
        self.subscriptions = set()
        self.protocol = None

        from twisted.internet import reactor
        self.client_endpoint = clientFromString(reactor, endpoint)
        self.client_factory = ClientFactory.forProtocol(_Protocol, ident, secret)
        self.client_factory.service = self

        self.client_service = ClientService(
            self.client_endpoint,
            self.client_factory,
            retryPolicy=retryPolicy,
        )
        self.client_service.setServiceParent(self)

        self.whenConnected = defer.Deferred()

    def publish(self, chan, payload):
        if self.protocol:
            self.protocol.publish(self.ident, chan, payload)

    def subscribe(self, topic):
        if topic not in self.subscriptions:
            self.subscriptions.add(topic)
            if self.protocol:
                self.protocol.subscribe(self.ident, topic)

    def unsubscribe(self, topic):
        if topic in self.subscriptions:
            self.subscriptions.discard(topic)
            if self.protocol:
                self.protocol.unsubscribe(self.ident, topic)

    def read(self):
        '''
        Returns a Deferred which fires with the next available message received
        from the broker.

        If a message has already been received and is queue the deferred will
        fire immediately.
        '''
        return self.read_queue.get()



import os
import unittest
from unittest import mock

from hpfeeds.broker.auth.env import Authenticator


class TestAuthenticator(unittest.TestCase):

    def test_1(self):
        environ = {
            'HPFEEDS_AAA_SECRET': 'bbb',
            'HPFEEDS_AAA_OWNER': 'ccc',
            'HPFEEDS_AAA_PUBCHANS': 'a,b,c',
            'HPFEEDS_AAA_SUBCHANS': 'd,e,f',
        }

        with mock.patch.dict(os.environ, environ):
            key = Authenticator().get_authkey('aaa')

        assert key['owner'] == 'ccc'
        assert key['secret'] == 'bbb'
        assert key['pubchans'] == ['a', 'b', 'c']
        assert key['subchans'] == ['d', 'e', 'f']

    def test_2(self):
        environ = {
            'HPFEEDS_AAA_SECRET': 'bbb',
            'HPFEEDS_AAA_PUBCHANS': 'a,b,c',
            'HPFEEDS_AAA_SUBCHANS': 'd,e,f',
        }

        with mock.patch.dict(os.environ, environ):
            key = Authenticator().get_authkey('aaa')

        assert key['owner'] == 'aaa'
        assert key['secret'] == 'bbb'
        assert key['pubchans'] == ['a', 'b', 'c']
        assert key['subchans'] == ['d', 'e', 'f']



import unittest
from unittest import mock

from hpfeeds.asyncio import BaseProtocol, ClientProtocol


class TestAioBaseProtocol(unittest.TestCase):

    def setUp(self):
        self.protocol = BaseProtocol()
        self.protocol.transport = mock.Mock()

    def patch_object(self, *args, **kwargs):
        patcher = mock.patch.object(*args, **kwargs)
        self.addCleanup(patcher.stop)
        return patcher.start()

    def test_on_error(self):
        self.assertRaises(NotImplementedError, self.protocol.on_error, 'error')

    def test_on_info(self):
        self.assertRaises(NotImplementedError, self.protocol.on_info, 'name', b'rand')

    def test_on_auth(self):
        self.assertRaises(NotImplementedError, self.protocol.on_auth, 'name', b'rand')

    def test_on_publish(self):
        self.assertRaises(NotImplementedError, self.protocol.on_publish, 'ident', 'chan', 'data')

    def test_on_subscribe(self):
        self.assertRaises(NotImplementedError, self.protocol.on_subscribe, 'ident', 'chan')

    def test_on_unsubscribe(self):
        self.assertRaises(NotImplementedError, self.protocol.on_unsubscribe, 'ident', 'chan')

    def test_error(self):
        self.protocol.error('error')
        assert self.protocol.transport.write.call_args[0][0] == b'\x00\x00\x00\n\x00error'

    def test_info(self):
        self.protocol.info('name', b'\x00' * 4)
        assert self.protocol.transport.write.call_args[0][0] == b'\x00\x00\x00\x0e\x01\x04name\x00\x00\x00\x00'

    def test_auth(self):
        self.protocol.auth(b'\x00' * 4, 'ident', 'secret')
        assert self.protocol.transport.write.call_args[0][0] == \
            b'\x00\x00\x00\x1f\x02\x05ident\x16\xa3\x11\xd5\xc2`\xcd\xc1\xee\xf3\x8b\xaf"\xdf\x97\x18\x90t&\xac'

    def test_publish(self):
        self.protocol.publish('ident', 'chan', 'payload')
        assert self.protocol.transport.write.call_args[0][0] == b'\x00\x00\x00\x17\x03\x05ident\x04chanpayload'

    def test_subscribe(self):
        self.protocol.subscribe('ident', 'chan')
        assert self.protocol.transport.write.call_args[0][0] == b'\x00\x00\x00\x0f\x04\x05identchan'

    def test_unsubscribe(self):
        self.protocol.unsubscribe('ident', 'chan')
        assert self.protocol.transport.write.call_args[0][0] == b'\x00\x00\x00\x0f\x05\x05identchan'

    def test_protocol_error(self):
        self.protocol.protocol_error('reason')


class TestAioBaseProtocolDecoding(unittest.TestCase):

    def setUp(self):
        self.protocol = BaseProtocol()
        self.on_error = self.patch_object(self.protocol, 'on_error')
        self.on_info = self.patch_object(self.protocol, 'on_info')
        self.on_auth = self.patch_object(self.protocol, 'on_auth')
        self.on_publish = self.patch_object(self.protocol, 'on_publish')
        self.on_subscribe = self.patch_object(self.protocol, 'on_subscribe')
        self.on_unsubscribe = self.patch_object(self.protocol, 'on_unsubscribe')
        self.protocol_error = self.patch_object(self.protocol, 'protocol_error')
        self.protocol.transport = mock.Mock()

    def patch_object(self, *args, **kwargs):
        patcher = mock.patch.object(*args, **kwargs)
        self.addCleanup(patcher.stop)
        return patcher.start()

    def test_on_error(self):
        self.protocol.data_received(b'\x00\x00\x00\n\x00error')
        assert self.on_error.call_args[0][0] == 'error'

    def test_on_info(self):
        self.protocol.data_received(b'\x00\x00\x00\x0e\x01\x04name\x00\x00\x00\x00')
        assert self.on_info.call_args[0][0] == 'name'
        assert self.on_info.call_args[0][1] == b'\x00' * 4

    def test_on_auth(self):
        self.protocol.data_received(b'\x00\x00\x00\x1f\x02\x05ident\x16\xa3\x11\xd5\xc2`\xcd\xc1\xee\xf3\x8b\xaf"\xdf\x97\x18\x90t&\xac')
        assert self.on_auth.call_args[0][0] == 'ident'
        assert self.on_auth.call_args[0][1] == b'\x16\xa3\x11\xd5\xc2`\xcd\xc1\xee\xf3\x8b\xaf"\xdf\x97\x18\x90t&\xac'

    def test_on_publish(self):
        self.protocol.data_received(b'\x00\x00\x00\x17\x03\x05ident\x04chanpayload')
        assert self.on_publish.call_args[0][0] == 'ident'
        assert self.on_publish.call_args[0][1] == 'chan'
        assert self.on_publish.call_args[0][2] == b'payload'

    def test_on_subscribe(self):
        self.protocol.data_received(b'\x00\x00\x00\x0f\x04\x05identchan')
        assert self.on_subscribe.call_args[0][0] == 'ident'
        assert self.on_subscribe.call_args[0][1] == 'chan'

    def test_on_unsubscribe(self):
        self.protocol.data_received(b'\x00\x00\x00\x0f\x05\x05identchan')
        assert self.on_unsubscribe.call_args[0][0] == 'ident'
        assert self.on_unsubscribe.call_args[0][1] == 'chan'

    def test_invalid_opcode(self):
        # We test this seperately as the unpacker also enforces valid opcodes
        # So normally this won't be hit.
        self.protocol.message_received(77, b'\x05identchan')
        assert self.protocol_error.call_args[0][0] == 'Unknown message opcode: 77'
        self.protocol.transport.close.assert_called_with()

    def test_invalid_opcode_2(self):
        self.protocol.data_received(b'\x00\x00\x00\x0f\x06\x05identchan')
        assert self.protocol_error.call_args[0][0] == 'Unknown opcode: 6'
        self.protocol.transport.close.assert_called_with()

    def test_invalid_size(self):
        self.protocol.data_received(b'\x00\xff\xff\xff\x05\x05identchan')
        assert self.protocol_error.call_args[0][0] == 'Message too big; op 5 ml: 16777215 max_ml: 1048576'
        self.protocol.transport.close.assert_called_with()


class TestAioClientProtocol(unittest.TestCase):

    def setUp(self):
        self.protocol = ClientProtocol('ident', 'secret')
        self.protocol.transport = mock.Mock()

        self.connection_ready = self.patch_object(self.protocol, 'connection_ready')
        self.protocol_error = self.patch_object(self.protocol, 'protocol_error')

    def patch_object(self, *args, **kwargs):
        patcher = mock.patch.object(*args, **kwargs)
        self.addCleanup(patcher.stop)
        return patcher.start()

    def test_connection_ready(self):
        self.protocol.connection_ready()

    def test_on_info(self):
        # A client should auto-reply to an OP_INFO the call connection_ready
        self.protocol.on_info('hpfeeds', b'\x00' * 4)
        assert self.protocol.transport.write.call_args[0][0] == \
            b'\x00\x00\x00\x1f\x02\x05ident\x16\xa3\x11\xd5\xc2`\xcd\xc1\xee\xf3\x8b\xaf"\xdf\x97\x18\x90t&\xac'

    def test_on_auth(self):
        ''' Client should never receive an OP_AUTH message - it is an error if it does '''
        self.protocol.data_received(b'\x00\x00\x00\x1f\x02\x05ident\x16\xa3\x11\xd5\xc2`\xcd\xc1\xee\xf3\x8b\xaf"\xdf\x97\x18\x90t&\xac')
        assert self.protocol_error.call_args[0][0] == 'Unexpected OP_AUTH'
        self.protocol.transport.close.assert_called_with()

    def test_on_subscribe(self):
        ''' Client should never receive an OP_SUBSCRIBE message - it is an error if it does '''
        self.protocol.data_received(b'\x00\x00\x00\x0f\x04\x05identchan')
        assert self.protocol_error.call_args[0][0] == 'Unexpected OP_SUBSCRIBE'
        self.protocol.transport.close.assert_called_with()

    def test_on_unsubscribe(self):
        ''' Client should never receive an OP_UNSUBSCRIBE message - it is an error if it does '''
        self.protocol.data_received(b'\x00\x00\x00\x0f\x05\x05identchan')
        assert self.protocol_error.call_args[0][0] == 'Unexpected OP_UNSUBSCRIBE'
        self.protocol.transport.close.assert_called_with()

    def test_error(self):
        self.assertRaises(RuntimeError, self.protocol.error, 'error')

    def test_info(self):
        self.assertRaises(RuntimeError, self.protocol.info, 'name', b'\x00' * 4)


class TestAioClientProtocolConnReady(unittest.TestCase):

    def test_connection_ready(self):
        self.protocol = ClientProtocol('ident', 'secret')
        self.protocol.connection_ready()



#!/usr/bin/env python
from setuptools import find_packages, setup

setup(
    name='hpfeeds3',
    version='0.9.2',
    description='Python implementation of the honeypot feeds broker',
    author='John Carr',
    author_email='john.carr@unrouted.co.uk',
    url='https://github.com/Jc2k/hpfeeds',
    license='GPL',
    packages=find_packages(exclude=['ez_setup']),
    include_package_data=True,
    zip_safe=False,
    entry_points='''
        [console_scripts]
        hpfeeds = hpfeeds.scripts.cli:main
        hpfeeds-broker = hpfeeds.scripts.broker:main
    ''',
    extra_require = {
        'test': [
            'flake8',
            'flake8-isort',
            'pytest',
            'pytest-cov',
            'codecov',
        ],
        'broker': [
            'aiorun',
            'aiohttp',
            'prometheus_client',
        ],
    },
)



import asyncio
import logging
import socket
import threading
import unittest

from hpfeeds import client
from hpfeeds.broker import prometheus
from hpfeeds.broker.auth.memory import Authenticator
from hpfeeds.broker.server import Server
from hpfeeds.protocol import readpublish


class TestClientIntegration(unittest.TestCase):

    log = logging.getLogger('hpfeeds.testserver')

    def _server_thread(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self.server_future = loop.create_future()

        async def inner():
            authenticator = Authenticator({
                'test': {
                    'secret': 'secret',
                    'subchans': ['test-chan'],
                    'pubchans': ['test-chan'],
                    'owner': 'some-owner',
                }
            })

            self.server = Server(authenticator, sock=self.sock)

            self.log.debug('Starting server')
            future = asyncio.ensure_future(self.server.serve_forever())

            self.log.debug('Awaiting test teardown')
            await self.server_future

            self.log.debug('Stopping test server')
            future.cancel()
            await future

        loop.run_until_complete(inner())

    def setUp(self):
        prometheus.reset()

        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_client_connections') == 0

        self.sock = sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('127.0.0.1', 0))
        self.port = sock.getsockname()[1]

        self.server_thread = threading.Thread(
            target=self._server_thread,
        )
        self.server_thread.start()

    def test_subscribe_and_publish(self):
        c = client.new('127.0.0.1', self.port, 'test', 'secret')

        c.subscribe('test-chan')
        c._subscribe()

        # If we have subscribed to a channel we should be able to see a
        # connection in monitoring
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_client_connections') == 1

        c.publish('test-chan', b'data')

        opcode, data = c._read_message()
        assert opcode == 3
        assert readpublish(data) == ('test', 'test-chan', b'data')

        # We managed to publish a message - check this is reflected in stats
        assert 1 == prometheus.REGISTRY.get_sample_value(
            'hpfeeds_broker_receive_publish_count',
            {'ident': 'test', 'chan': 'test-chan'}
        )

        # If we managed to read a message from the broker then we must be subscribed
        # Check this is reflected in stats
        assert 1 == prometheus.REGISTRY.get_sample_value(
            'hpfeeds_broker_subscriptions',
            {'ident': 'test', 'chan': 'test-chan'}
        )

        self.log.debug('Stopping client')
        c.stop()

        self.log.debug('Closing client')
        c.close()

    def tearDown(self):
        self.log.debug('Cancelling future')
        self.server_future.set_result(None)
        self.log.debug('Waiting')
        self.server_thread.join()

        assert len(self.server.connections) == 0, 'Connection left dangling'
        assert prometheus.REGISTRY.get_sample_value('hpfeeds_broker_client_connections') == 0

        assert 0 == prometheus.REGISTRY.get_sample_value(
            'hpfeeds_broker_subscriptions',
            {'ident': 'test', 'chan': 'test-chan'}
        )



#!/usr/bin/python
# -*- coding: utf8 -*-

import logging
import os
import socket
import sys

from hpfeeds.asyncio.protocol import BaseProtocol
from hpfeeds.protocol import OP_AUTH, hashsecret

from .prometheus import (
    CLIENT_CONNECTIONS,
    CONNECTION_ERROR,
    CONNECTION_LOST,
    CONNECTION_MADE,
    CONNECTION_READY,
)

log = logging.getLogger('hpfeeds.broker.connection')


class Connection(BaseProtocol):

    def __init__(self, server):
        self.server = server

        self.uid = None
        self.ak = None
        self.pubchans = []
        self.subchans = []

        self.active_subscriptions = set()

        self.authrand = os.urandom(4)

        super().__init__()

    def connection_made(self, transport):
        CLIENT_CONNECTIONS.inc()
        CONNECTION_MADE.inc()

        self.server.connections.add(self)

        self.transport = transport
        self.peer, self.port = transport.get_extra_info('peername')

        sock = transport.get_extra_info('socket')
        if sock is not None:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            if sys.platform.startswith('linux'):
                sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 10)
                sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 5)
                sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 3)

        self.info(self.server.name, self.authrand)

        log.debug(f'{self}: Sent auth challenge')

    def connection_lost(self, reason):
        CLIENT_CONNECTIONS.dec()
        CONNECTION_LOST.labels(self.ak).inc()

        for chan in list(self.active_subscriptions):
            self.server.unsubscribe(self, chan)

        self.server.connections.discard(self)
        self.server = None

        log.debug(f'Disconnection from {self}; cleanup completed.')

    def __str__(self):
        peer, port = self.peer, self.port
        ident = self.ak
        owner = self.uid
        return (
            f'<Connection ident={ident} owner={owner} peer={peer} port={port}'
        )

    def message_received(self, opcode, message):
        if not self.uid and opcode != OP_AUTH:
            CONNECTION_ERROR.labels('', 'invalid-first-message').inc()
            self.error("First message was not AUTH")
            self.transport.close()
            return

        return super().message_received(opcode, message)

    def on_auth(self, ident, secret):
        akrow = self.server.get_authkey(ident)
        if not akrow:
            CONNECTION_ERROR.labels(ident, 'invalid-ident').inc()
            self.error(f"Authentication failed for {ident}")
            self.transport.close()
            return

        akhash = hashsecret(self.authrand, akrow["secret"])
        if not akhash == secret:
            CONNECTION_ERROR.labels(ident, 'invalid-secret').inc()
            self.error(f"Authentication failed for {ident}")
            self.transport.close()
            return

        self.ak = ident
        self.uid = akrow["owner"]
        self.pubchans = akrow.get("pubchans", [])
        self.subchans = akrow.get("subchans", [])

        CONNECTION_READY.labels(ident).inc()

    def on_publish(self, ident, chan, payload):
        if not ident == self.ak:
            CONNECTION_ERROR.labels(ident, 'ident-mismatch-pub').inc()
            self.error(f"Invalid authkey in message, ident={ident}")
            self.transport.close()
            return

        if chan not in self.pubchans:
            CONNECTION_ERROR.labels(ident, 'no-pub-permission').inc()
            self.error(f'Authkey not allowed to pub here. ident={ident}, chan={chan}')
            self.transport.close()
            return

        self.server.publish(self, chan, payload)

    def on_subscribe(self, ident, chan):
        if chan not in self.subchans:
            CONNECTION_ERROR.labels(ident, 'no-sub-permission').inc()
            self.error(f'Authkey not allowed to sub here. ident={self.ak}, chan={chan}')
            self.transport.close()
        self.server.subscribe(self, chan)

    def on_unsubscribe(self, ident, chan):
        self.server.unsubscribe(self, chan)



import unittest
from unittest import mock

from hpfeeds.broker import prometheus
from hpfeeds.broker.auth.memory import Authenticator
from hpfeeds.broker.connection import Connection
from hpfeeds.broker.server import Server
from hpfeeds.protocol import (
    Unpacker,
    msgauth,
    msgpublish,
    msgsubscribe,
    msgunsubscribe,
    readinfo,
    readpublish,
)


def parse(mock_write):
    unpacker = Unpacker()
    for call in mock_write.call_args_list:
        unpacker.feed(call[0][0])
    results = []
    for msg in unpacker:
        results.append(msg)
    return results


class TestBrokerConnection(unittest.TestCase):

    def setUp(self):
        prometheus.reset()

        authenticator = Authenticator({
            'test': {
                'secret': 'secret',
                'subchans': ['test-chan'],
                'pubchans': ['test-chan'],
                'owner': 'some-owner',
            }
        })

        self.server = Server(authenticator, bind='127.0.0.1:20000')

    def make_connection(self):
        transport = mock.Mock()
        transport.get_extra_info.side_effect = lambda name: ('127.0.0.1', 80) if name == 'peername' else None

        connection = Connection(self.server)
        connection.connection_made(transport)

        return connection

    def test_sends_challenge(self):
        c = self.make_connection()
        assert parse(c.transport.write)[0][1][:-4] == b'\x07hpfeeds'

    def test_must_auth(self):
        c = self.make_connection()
        c.data_received(msgpublish('a', 'b', b'c'))

        assert parse(c.transport.write)[0][1][:-4] == b'\x07hpfeeds'
        assert parse(c.transport.write)[1][1] == b'First message was not AUTH'

    def test_auth_failure_wrong_secret(self):
        c = self.make_connection()
        name, rand = readinfo(parse(c.transport.write)[0][1])
        c.data_received(msgauth(rand, 'test', 'secret2'))

        assert parse(c.transport.write)[1][1] == b'Authentication failed for test'

    def test_auth_failure_no_such_ident(self):
        c = self.make_connection()
        name, rand = readinfo(parse(c.transport.write)[0][1])
        c.data_received(msgauth(rand, 'test2', 'secret'))

        assert parse(c.transport.write)[1][1] == b'Authentication failed for test2'

    def test_permission_to_sub(self):
        c = self.make_connection()
        name, rand = readinfo(parse(c.transport.write)[0][1])
        c.data_received(msgauth(rand, 'test', 'secret'))
        c.data_received(msgsubscribe('test', 'test-chan2'))

        assert parse(c.transport.write)[1][1] == b'Authkey not allowed to sub here. ident=test, chan=test-chan2'

    def test_permission_to_pub(self):
        c = self.make_connection()
        name, rand = readinfo(parse(c.transport.write)[0][1])
        c.data_received(msgauth(rand, 'test', 'secret'))
        c.data_received(msgpublish('test', 'test-chan2', b'c'))

        assert parse(c.transport.write)[1][1] == b'Authkey not allowed to pub here. ident=test, chan=test-chan2'

    def test_pub_ident_checked(self):
        c = self.make_connection()
        name, rand = readinfo(parse(c.transport.write)[0][1])
        c.data_received(msgauth(rand, 'test', 'secret'))
        c.data_received(msgpublish('wrong-ident', 'test-chan2', b'c'))

        assert parse(c.transport.write)[1][1] == b'Invalid authkey in message, ident=wrong-ident'

    def test_auth_success(self):
        c = self.make_connection()
        name, rand = readinfo(parse(c.transport.write)[0][1])
        c.data_received(msgauth(rand, 'test', 'secret'))
        c.data_received(msgsubscribe('test', 'test-chan'))
        c.data_received(msgpublish('test', 'test-chan', b'c'))

        assert readpublish(parse(c.transport.write)[1][1]) == (
            'test',
            'test-chan',
            b'c'
        )

    def test_multiple_subscribers(self):
        subscribers = []
        for i in range(5):
            c = self.make_connection()
            name, rand = readinfo(parse(c.transport.write)[0][1])
            c.data_received(msgauth(rand, 'test', 'secret'))
            c.data_received(msgsubscribe('test', 'test-chan'))
            subscribers.append(c)

        c = self.make_connection()
        name, rand = readinfo(parse(c.transport.write)[0][1])
        c.data_received(msgauth(rand, 'test', 'secret'))
        c.data_received(msgpublish('test', 'test-chan', b'c'))

        for c in subscribers:
            msgs = parse(c.transport.write)
            assert readpublish(msgs[1][1]) == (
                'test',
                'test-chan',
                b'c'
            )

    def test_auth_unsubscribe(self):
        c = self.make_connection()
        name, rand = readinfo(parse(c.transport.write)[0][1])
        c.data_received(msgauth(rand, 'test', 'secret'))

        c.data_received(msgsubscribe('test', 'test-chan'))
        c.data_received(msgpublish('test', 'test-chan', b'c'))
        c.data_received(msgunsubscribe('test', 'test-chan'))
        c.data_received(msgpublish('test', 'test-chan', b'c'))
        c.data_received(msgsubscribe('test', 'test-chan'))
        c.data_received(msgpublish('test', 'test-chan', b'c'))

        messages = parse(c.transport.write)
        for msg in messages[1:]:
            assert readpublish(msg[1]) == (
                'test',
                'test-chan',
                b'c'
            )

        # 1 auth and 2 publish
        assert len(messages) == 3



class FeedException(Exception):
    pass


class Disconnect(Exception):
    pass


class ProtocolException(Disconnect):
    pass


class MessageTooBig(ProtocolException):

    def __init__(self, op, ml, max_ml):
        super(MessageTooBig, self).__init__(
            'Message too big; op {op} ml: {ml} max_ml: {max_ml}'.format(
                op=op,
                ml=ml,
                max_ml=max_ml,
            )
        )


class BadClient(ProtocolException):
    pass



from aiohttp import web
from prometheus_client import (
    REGISTRY,
    Counter,
    Gauge,
    Histogram,
    generate_latest,
)

CLIENT_CONNECTIONS = Gauge(
    'hpfeeds_broker_client_connections',
    'Number of clients connected to broker',
)

CONNECTION_MADE = Counter(
    'hpfeeds_broker_connection_made',
    'Number of connections established',
)

CONNECTION_READY = Counter(
    'hpfeeds_broker_connection_ready',
    'Number of connections established + authenticated',
    ['ident'],
)

CONNECTION_ERROR = Counter(
    'hpfeeds_broker_connection_error',
    'Number of connections that experienced a protocol error',
    ['ident', 'category'],
)

CONNECTION_LOST = Counter(
    'hpfeeds_broker_connection_lost',
    'Number of connections lost',
    ['ident'],
)

SUBSCRIPTIONS = Gauge(
    'hpfeeds_broker_subscriptions',
    'Number of subscriptions to a channel',
    ['ident', 'chan'],
)

RECEIVE_PUBLISH_COUNT = Counter(
    'hpfeeds_broker_receive_publish_count',
    'Number of events received by broker for a channel',
    ['ident', 'chan'],
)

RECEIVE_PUBLISH_SIZE = Histogram(
    'hpfeeds_broker_receive_publish_size',
    'Sizes of messages received by broker for a channel',
    ['ident', 'chan'],
)


def reset():
    ''' Reset the metrics to 0. This is intended for tests **only**. '''
    CLIENT_CONNECTIONS._value.set(0)
    SUBSCRIPTIONS._metrics = {}
    RECEIVE_PUBLISH_SIZE._metrics = {}
    RECEIVE_PUBLISH_COUNT._metrics = {}
    CONNECTION_ERROR._metrics = {}
    CONNECTION_LOST._metrics = {}
    CONNECTION_MADE._value.set(0)
    CONNECTION_READY._metrics = {}


async def metrics(request):
    data = generate_latest(REGISTRY)
    return web.Response(text=data.decode('utf-8'), content_type='text/plain', charset='utf-8')


async def start_metrics_server(host, port):
    app = web.Application()
    app.router.add_get('/metrics', metrics)

    runner = web.AppRunner(app, access_log=None)
    await runner.setup()

    site = web.TCPSite(runner, host, port)

    await site.start()

    return runner



# Copyright (C) 2010-2013 Mark Schloesser <ms@mwcollect.org
# This file is part of hpfeeds - https://github.com/rep/hpfeeds
# See the file 'LICENSE' for copying permission.

import hashlib
import struct

from .exceptions import MessageTooBig, ProtocolException

BUFSIZ = 16384

OP_ERROR = 0
OP_INFO = 1
OP_AUTH = 2
OP_PUBLISH = 3
OP_SUBSCRIBE = 4
OP_UNSUBSCRIBE = 5

MAXBUF = 1024**2

SIZES = {
    OP_ERROR: 5 + MAXBUF,
    OP_INFO: 5 + 256 + 20,
    OP_AUTH: 5 + 256 + 20,
    OP_PUBLISH: 5 + MAXBUF,
}


def hashsecret(rand, secret):
    return hashlib.sha1(bytes(rand) + secret.encode('utf-8')).digest()


def force_bytes(value):
    if isinstance(value, str):
        return value.encode('utf-8')
    return value


def force_str(value):
    if isinstance(value, bytearray):
        value = bytes(value)
    if isinstance(value, bytes):
        return value.decode('utf-8')
    return value


def strpack8(x):
    # packs a string with 1 byte length field
    x = force_bytes(x)
    return struct.pack('!B', len(x)) + x


def strunpack8(x):
    # unpacks a string with 1 byte length field
    length = ord(x[0:1])
    return force_str(x[1:1+length]), x[1+length:]


def msghdr(op, data):
    return struct.pack('!iB', 5 + len(data), op) + data


def msginfo(name, rand):
    return msghdr(OP_INFO, strpack8(name) + force_bytes(rand))


def msgsubscribe(ident, chan):
    return msghdr(OP_SUBSCRIBE, strpack8(ident) + force_bytes(chan))


def msgunsubscribe(ident, chan):
    return msghdr(OP_UNSUBSCRIBE, strpack8(ident) + force_bytes(chan))


def msgpublish(ident, chan, data):
    return msghdr(
        OP_PUBLISH,
        strpack8(ident) + strpack8(chan) + force_bytes(data),
    )


def msgauth(rand, ident, secret):
    return msghdr(OP_AUTH, strpack8(ident) + hashsecret(rand, secret))


def msgerror(error):
    return msghdr(OP_ERROR, force_bytes(error))


def readinfo(data):
    ident, rand = strunpack8(data)
    return force_str(ident), rand


def readauth(data):
    ident, secret = strunpack8(data)
    return force_str(ident), secret


def readsubscribe(data):
    ident, rest = strunpack8(data)
    return force_str(ident), force_str(rest)


def readunsubscribe(data):
    ident, rest = strunpack8(data)
    return force_str(ident), force_str(rest)


def readpublish(data):
    ident, rest = strunpack8(data)
    chan, rest = strunpack8(rest)
    return force_str(ident), force_str(chan), rest


def readerror(error):
    return force_str(error)


class Unpacker(object):

    def __init__(self):
        self.reset()

    def __iter__(self):
        return self

    def __next__(self):
        return self.unpack()

    def next(self):
        # For python2.7 compatibility only
        return self.__next__()

    def reset(self):
        self.buf = bytearray()

    def feed(self, data):
        self.buf.extend(data)

    def ready(self):
        if len(self.buf) < 5:
            return False

        ml, opcode = struct.unpack('!iB', self.buf[0:5])

        if opcode < OP_ERROR or opcode > OP_UNSUBSCRIBE:
            raise ProtocolException('Unknown opcode: {}'.format(opcode))

        max_ml = SIZES.get(opcode, MAXBUF)
        if ml > max_ml:
            raise MessageTooBig(opcode, ml, max_ml)

        if len(self.buf) < ml:
            return False

        return True

    def pop(self):
        ml, opcode = struct.unpack('!iB', self.buf[0:5])
        data = bytes(self.buf[5:][:ml-5])
        del self.buf[:ml]
        return opcode, data

    def unpack(self):
        if not self.ready():
            raise StopIteration('No message')
        return self.pop()
